import asyncio
import logging
import signal
import sys

from .utils import current_origin

log = logging.getLogger(__name__)


loops = {}


def get_loop():
    # Here we get the magic cookie of our current thread in current pid
    origin = current_origin()
    if origin not in loops:
        loops[origin] = asyncio.new_event_loop()
        # asyncio.set_event_loop(loops[origin])
    return loops[origin]


def close_loop():
    origin = current_origin()
    try:
        loops[origin].close()
    finally:
        del loops[origin]


def clean_loops():
    log.info(f'Cleaning at exit {len(loops)} event loops')
    if not loops:
        return
    for loop in loops.values():
        loop.close()


def stop():
    for task in asyncio.Task.all_tasks():
        task.cancel()


def run(coro):
    loop = get_loop()
    try:
        loop.run_until_complete(coro)
    except asyncio.exceptions.CancelledError:
        from .websockets import die

        log.info("Loop got cancelled")
        die()
        sys.exit(0)

    if sys.platform != "win32":
        loop.add_signal_handler(signal.SIGTERM, stop)



import asyncio
import logging
import os
import webbrowser
from threading import Lock

from aiohttp import ClientSession
from aiohttp.client_exceptions import ClientConnectorError

from . import config
from .errors import NoServerFoundError
from .forking import forkserver
from .loops import get_loop
from .utils import current_origin

log = logging.getLogger(__name__)
fork_lock = Lock()

websockets = {}
sessions = {}

websocket_options = {"max_msg_size": 1024 * 1024 * 1024}


async def websocket():
    # Here we get the magic cookie of our current thread in current pid
    origin = current_origin()
    if origin in websockets:
        return websockets[origin]

    def url(side):
        overriden_port = config.front_port if side == 'front' else config.port
        protocol = config.protocol
        host = config.host
        return f'{protocol}://{host}:{overriden_port}/{side}/{origin}'

    sessions[origin] = sessions.get(origin, ClientSession())

    with fork_lock:
        try:
            ws = await sessions[origin].ws_connect(
                url('back'), **websocket_options
            )
            log.info('Found existing kalong server')
        except ClientConnectorError:
            # If there are no server available, fork one
            log.info('No kalong server, starting one')
            forkserver()
            for delay in [0.01, 0.025, 0.05, 0.1, 0.25, 0.5, 1, 2, 5, 10]:
                await asyncio.sleep(delay, loop=get_loop())
                try:
                    ws = await sessions[origin].ws_connect(
                        url('back'), **websocket_options
                    )
                    break
                except ClientConnectorError:
                    pass
            else:
                raise NoServerFoundError()

        # webbrowser.open should be in the mutex too, it's not thread safe
        if os.getenv('KALONG_NO_BROWSER') or not webbrowser.open(url('front')):
            log.warn(
                'Please open your browser to the following url: '
                f'{url("front")}'
            )

    websockets[origin] = ws
    return ws


async def close_websocket():
    origin = current_origin()
    try:
        if origin in websockets:
            await websockets[origin].close()
    finally:
        if origin in websockets:
            del websockets[origin]
        if origin in sessions:
            try:
                await sessions[origin].close()
            finally:
                del sessions[origin]


def die():
    log.info('Dying, closing socket.')
    loop = get_loop()
    if loop.is_running():
        loop.create_task(close_websocket())
    else:
        loop.run_until_complete(close_websocket())


def close_all(closeables):
    if not closeables:
        return
    loop = get_loop()
    loop.run_until_complete(
        asyncio.gather(
            *[closeable.close() for closeable in closeables.values()],
            loop=loop,
            return_exceptions=True,
        )
    )


def clean_websockets():
    log.info(
        f'Cleaning at exit {len(websockets)} ws and {len(sessions)} sessions'
    )
    close_all(websockets)
    close_all(sessions)



class KalongException(Exception):
    pass


class NoClientFoundError(KalongException):
    pass


class NoServerFoundError(KalongException):
    pass


class CantStartServerError(KalongException):
    pass



import logging
import os
from argparse import REMAINDER, ArgumentParser

defaults = {
    'server': False,
    'protocol': 'http',
    'host': 'localhost',
    'port': 59999,
    'front_port': 59999,
    'log': 'INFO',
    'detached': False,
    'command': [],
    'inject': None,
}


class Config:
    def __init__(self):
        self.__dict__.update(defaults)

    def get_parser(self):
        parser = ArgumentParser(description='Kalong cli')
        parser.add_argument(
            '--server',
            action='store_true',
            help='Launch the kalong server. '
            'This option is used by kalong itself',
        )
        parser.add_argument(
            '--protocol',
            default=self.protocol,
            help='Protocol for contacting kalong server',
        )
        parser.add_argument(
            '--host', default=self.host, help='Host of kalong server'
        )
        parser.add_argument(
            '--port', type=int, default=self.port, help='Port of kalong server'
        )
        front_port_default = (
            {'default': self.port} if self.port != self.front_port else {}
        )
        parser.add_argument(
            '--front-port',
            type=int,
            help='Port of kalong frontend, defaults to port option',
            **front_port_default,
        )
        parser.add_argument(
            '--log',
            choices=['critical', 'error', 'warning', 'info', 'debug'],
            default=self.log,
            help='Kalong verbosity',
        )
        parser.add_argument(
            '--detached',
            action='store_true',
            help="Server won't exit in this mode after last client disconnect",
        )
        parser.add_argument(
            '--inject',
            type=int,
            help="Pid of a running process in which a debugger will be "
            "injected with gdb. This needs a working gdb and ptrace enabled",
        )
        parser.add_argument(
            'command',
            nargs=REMAINDER,
            help='A python file to trace with kalong and its arguments or '
            'nothing to launch a shell instead',
        )
        return parser

    def parse_args(self):
        args = self.get_parser().parse_args()
        if not args.front_port:
            args.front_port = args.port
        return args

    def from_args(self):
        for name, arg in vars(self.parse_args()).items():
            setattr(self, name, arg)

    def from_env(self):
        for name, value in defaults.items():
            value = os.getenv(f'KALONG_{name.upper()}')
            if value:
                setattr(self, name, type(name)(value))

    def get_args_for_server(self):
        yield '--server'

        if self.detached:
            yield '--detached'

        for name, default in defaults.items():
            if name in ['server', 'detached', 'command']:
                continue
            value = getattr(self, name)
            if value != default:
                pretty_name = name.replace('_', '-')
                yield f'--{pretty_name}={value}'

    @property
    def log_level(self):
        return getattr(logging, self.log.upper())

    def __str__(self):
        return str(self.__dict__)



import ctypes
from inspect import (
    getcomments,
    getmodule,
    getmro,
    getsourcefile,
    getsourcelines,
    iscode,
    signature,
)

from .doc_lookup import get_fqn, get_online_doc


class ObjCache(object):
    def __init__(self):
        self.cache = {}

    def register(self, obj):
        ident = id(obj)
        self.cache[ident] = obj
        return ident

    def get(self, ident):
        return self.cache[ident]

    def clear(self):
        self.cache = {}


obj_cache = ObjCache()


def walk_iterable(iterable, walked):
    return [walk_obj(obj, walked) for obj in iterable]


def walk_mapping(mapping, walked):
    return [
        {'key': walk_obj(key, walked), 'value': walk_obj(val, walked)}
        for key, val in mapping.items()
    ]


def walk_obj(obj, walked):
    id = obj_cache.register(obj)
    if id in walked:
        # Don't walk circular dependencies
        return {'type': 'obj', 'value': repr(obj), 'id': id}
    walked.add(id)

    if any(
        [isinstance(obj, list), isinstance(obj, set), isinstance(obj, tuple)]
    ):
        return {
            'type': 'iterable',
            'subtype': type(obj).__name__,
            'values': walk_iterable(obj, walked),
            'id': obj_cache.register(obj),
        }
    if isinstance(obj, dict):
        return {
            'type': 'mapping',
            'subtype': type(obj).__name__,
            'values': walk_mapping(obj, walked),
            'id': obj_cache.register(obj),
        }
    return {'type': 'obj', 'value': repr(obj), 'id': id}


def get_bases(cls):
    return {
        'id': obj_cache.register(cls),
        'name': cls.__name__,
        'bases': [get_bases(base) for base in cls.__bases__],
    }


def has_mixins(bases):
    return len(bases['bases']) > 1 or any(
        has_mixins(base) for base in bases['bases']
    )


def get_infos(obj):
    infos = {}
    infos['type'] = type(obj).__name__
    cls = obj if type(obj) == type else type(obj)
    try:
        infos['signature'] = obj.__name__ + str(signature(obj))
    except Exception:
        pass

    infos['fqn'] = get_fqn(obj)
    if infos['fqn']:
        infos['online_doc'] = get_online_doc(infos['fqn'])
    elif type(obj) != type:
        type_fqn = get_fqn(type(obj))
        if type_fqn:
            infos['online_doc'] = get_online_doc(type_fqn)

    try:
        infos['bases'] = get_bases(cls)
        if has_mixins(infos['bases']):
            try:
                infos['mro'] = [
                    {'id': obj_cache.register(m), 'type': m.__name__}
                    for m in (getmro(cls))
                ]
            except Exception:
                pass
    except Exception:
        pass

    try:
        infos['file'] = getsourcefile(obj)
    except Exception:
        infos['file'] = ''

    try:
        srcs, lno = getsourcelines(obj)
        infos['file'] += f':{lno}'
        infos['source_size'] = len(srcs)
    except Exception:
        pass

    try:
        infos['module'] += f':{getmodule(obj).__name__}'
    except Exception:
        pass

    infos['comments'] = getcomments(obj)

    infos['id'] = id(obj)

    return infos


def sync_locals(frame, f_locals):
    """
        This is a cpython hack to synchronize new locals back into
        the fast local slots.
    """
    try:
        frame.f_locals.clear()
        ctypes.pythonapi.PyFrame_LocalsToFast(
            ctypes.py_object(frame), ctypes.c_int(1)
        )
        frame.f_locals.update(**f_locals)
        ctypes.pythonapi.PyFrame_LocalsToFast(
            ctypes.py_object(frame), ctypes.c_int(0)
        )
    except Exception:
        pass


def get_code(obj):
    if iscode(obj):
        return obj
    if hasattr(obj, '__func__'):
        obj = obj.__func__
    for attr in ['__code__', 'gi_code', 'ag_code', 'cr_code', 'f_code']:
        if hasattr(obj, attr):
            code = getattr(obj, attr)
            if hasattr(code, 'co_code'):
                return code


def safe_getattr(obj, key, default):
    try:
        return getattr(obj, key)
    except Exception:
        return default



import difflib
import dis
import linecache
import logging
import os
import sys
import time
from inspect import (
    getdoc,
    getsource,
    isclass,
    iscoroutine,
    isgenerator,
    ismodule,
    isroutine,
    signature,
)
from itertools import groupby
from pathlib import Path
from pprint import pformat

from jedi import Interpreter

from .utils.io import capture_display, capture_exception, capture_std
from .utils.iterators import iter_cause, iter_stack, iter_frame
from .utils.obj import (
    get_code,
    get_infos,
    obj_cache,
    safe_getattr,
    sync_locals,
)

try:
    from cutter import cut
    from cutter.utils import bang_compile as compile
except ImportError:
    cut = None

try:
    from uncompyle6 import code_deparse
except Exception:
    code_deparse = None

log = logging.getLogger(__name__)


def get_frame(frame, key):
    if not key:
        return frame

    for f in iter_frame(frame):
        if id(f.f_code) == key:
            return f
    log.warn(f"Frame {key} not found")
    return frame


def serialize_frames(current_frame, current_tb):
    for frame, lno in iter_stack(current_frame, current_tb):
        code = frame.f_code
        filename = code.co_filename or '<unspecified>'
        if filename == '<frozen importlib._bootstrap>':
            filename = os.path.join(
                os.path.dirname(linecache.__file__),
                'importlib',
                '_bootstrap.py',
            )
        line = None
        linecache.checkcache(filename)
        line = linecache.getline(filename, lno, frame.f_globals)
        line = line and line.strip()
        startlnos = dis.findlinestarts(code)
        lastlineno = list(startlnos)[-1][1]
        fn = Path(filename).resolve()
        yield {
            'key': id(code),
            'absoluteFilename': str(fn),
            'filename': fn.name,
            'function': code.co_name,
            'firstFunctionLineNumber': code.co_firstlineno,
            'lastFunctionLineNumber': lastlineno,
            'lineNumber': lno,
            'lineSource': line,
            'active': frame == current_frame,
        }


def serialize_answer(prompt, frame):
    prompt = prompt.strip()
    duration = 0
    answer = []
    f_locals = dict(frame.f_locals)
    f_locals["_current_frame"] = frame
    with capture_exception(answer), capture_display(answer), capture_std(
        answer
    ):
        compiled_code = None
        try:
            compiled_code = compile(prompt, '<stdin>', 'single')
        except Exception:
            try:
                compiled_code = compile(prompt, '<stdin>', 'exec')
            except Exception:
                # handle ex
                sys.excepthook(*sys.exc_info())

        start = time.time()
        if compiled_code is not None:
            try:
                exec(compiled_code, frame.f_globals, f_locals)
            except Exception:
                # handle ex
                sys.excepthook(*sys.exc_info())
            del f_locals["_current_frame"]
            sync_locals(frame, f_locals)
        duration = int((time.time() - start) * 1000 * 1000 * 1000)

    return {'prompt': prompt, 'answer': answer, 'duration': duration}


def serialize_exception(type_, value, tb):
    return {
        'type': 'exception',
        'id': obj_cache.register(value),
        'subtype': 'root',
        'name': type_.__name__,
        'description': str(value),
        'traceback': list(serialize_frames(None, tb)),
        'causes': [
            {
                'id': obj_cache.register(cause),
                'subtype': 'cause' if explicit else 'context',
                'name': type(cause).__name__,
                'description': str(cause),
                'traceback': list(serialize_frames(None, cause.__traceback__)),
            }
            for cause, explicit in iter_cause(value)
        ],
    }


def attribute_classifier(attr):
    key = attr['key']
    value = attr['value']
    if key.startswith('__') and key.endswith('__'):
        return '__core__'
    if ismodule(value):
        return 'module'
    if isclass(value):
        return 'class'
    if isroutine(value):
        return 'method'
    # if isfunction(value):
    #     return 'function'
    if isgenerator(value):
        return 'generator'
    if iscoroutine(value):
        return 'coroutine'
    return 'attribute'


def serialize_attribute(attr, group):
    if group in ['function', 'method', 'coroutine']:
        try:
            attr['signature'] = str(signature(attr['value']))
        except Exception:
            pass
    attr['value'] = repr(attr['value'])
    return attr


def get_id_from_expression(expr, frame):
    return obj_cache.register(eval(expr, frame.f_globals, frame.f_locals))


def serialize_inspect_eval(prompt, frame):
    try:
        key = get_id_from_expression(prompt, frame)
    except Exception:
        return {
            'prompt': prompt,
            'answer': [serialize_exception(*sys.exc_info())],
        }
    return serialize_inspect(key, frame)


def serialize_inspect(key):
    obj = obj_cache.get(key)
    attributes = [
        {'key': key, 'value': value, 'id': obj_cache.register(value)}
        for key, value in [
            (
                key,
                safe_getattr(
                    obj,
                    key,
                    f'?! Broken obj: key {key!r} is listed in dir() but '
                    + f'getattr(obj, {key!r}) raises.',
                ),
            )
            for key in dir(obj)
        ]
    ]

    grouped_attributes = {
        group: [serialize_attribute(attr, group) for attr in attrs]
        for group, attrs in groupby(
            sorted(attributes, key=attribute_classifier),
            key=attribute_classifier,
        )
    }
    infos = get_infos(obj)
    doc = getdoc(obj)
    source = None
    try:
        source = getsource(obj)
    except Exception:
        if code_deparse:
            try:
                code = get_code(obj)
                if code:
                    uncompiled = code_deparse(code).text
                    source = f'# Decompiled from {obj!r}\n\n{uncompiled}'
            except Exception:
                raise  # TODO REMOVE

    answer = [
        {
            'type': 'inspect',
            'infos': infos,
            'attributes': grouped_attributes,
            'doc': doc,
            'source': source,
        }
    ]

    return {'prompt': repr(obj), 'answer': answer}


def serialize_diff_eval(leftStr, rightStr, frame):
    try:
        leftKey = get_id_from_expression(leftStr, frame)
    except Exception:
        return {
            'prompt': leftStr,
            'answer': [serialize_exception(*sys.exc_info())],
        }

    try:
        rightKey = get_id_from_expression(rightStr, frame)
    except Exception:
        return {
            'prompt': rightStr,
            'answer': [serialize_exception(*sys.exc_info())],
        }

    left = obj_cache.get(leftKey)
    right = obj_cache.get(rightKey)

    answer = [
        {
            'type': 'diff',
            'diff': ''.join(
                difflib.unified_diff(
                    (pformat(left, indent=2, width=5) + '\n').splitlines(
                        keepends=True
                    ),
                    (pformat(right, indent=2, width=5) + '\n').splitlines(
                        keepends=True
                    ),
                    fromfile=leftStr,
                    tofile=rightStr,
                )
            ),
        }
    ]

    return {'prompt': f'{leftStr} ? {rightStr}', 'answer': answer}


def serialize_suggestion(prompt, from_, to, cursor, frame):
    answer = {'prompt': prompt, 'from': from_, 'to': to, 'suggestion': {}}
    try:
        script = Interpreter(
            prompt,
            [frame.f_locals, frame.f_globals],
            line=cursor['line'] + 1,
            column=cursor['ch'],
        )
        completions = script.completions()
    except Exception:
        log.exception('Completion failed')
        return answer

    params_first_completions = [
        c for c in completions if c.name_with_symbols != c.name
    ] + [c for c in completions if c.name_with_symbols == c.name]

    completions = [
        {
            'text': comp.name_with_symbols,
            'description': comp.description,
            'base': comp.name_with_symbols[
                : len(comp.name_with_symbols) - len(comp.complete)
            ],
            'complete': comp.complete,
        }
        for comp in params_first_completions
    ]

    suggestion = {'from': from_, 'to': to, 'list': completions}
    answer['suggestion'] = suggestion
    return answer



import json
import linecache
import logging

from aiohttp import WSMsgType

from . import config
from .debugger import (
    serialize_answer,
    serialize_diff_eval,
    serialize_frames,
    serialize_inspect,
    serialize_inspect_eval,
    serialize_suggestion,
    get_frame,
)
from .loops import run
from .stepping import add_step, clear_step, stop_trace
from .utils import basicConfig
from .websockets import websocket, die

log = logging.getLogger(__name__)
basicConfig(level=config.log_level)


def communicate(frame, tb=None):
    run(communication_loop(frame, tb))


def initiate(event, frame, arg):
    if event == 'line':
        title = 'Tracing'
    elif event == 'call':
        title = f'Calling {frame.f_code.co_name}'
    elif event == 'return':
        title = f'Returning from {frame.f_code.co_name}'
    elif event == 'exception':
        title = f'{arg[0].__name__}: {arg[1]}'
    elif event == 'shell':
        title = 'Shell'
    run(
        send_once(
            {'type': 'PAUSE'},
            {'type': 'SET_THEME', 'theme': event},
            {'type': 'SET_TITLE', 'title': title},
            {
                'type': 'SET_FRAMES',
                'frames': list(
                    serialize_frames(
                        frame, arg[2] if event == "exception" else None
                    )
                )
                if event != 'shell'
                else [],
            },
        )
    )


async def send_once(*msgs):
    ws = await websocket()
    for msg in msgs:
        await ws.send_json(msg)


async def communication_loop(frame, tb=None):
    ws = await websocket()
    stop = False
    async for msg in ws:
        if msg.type == WSMsgType.TEXT:
            data = json.loads(msg.data)
            if data['type'] == 'GET_FRAMES':
                response = {
                    'type': 'SET_FRAMES',
                    'frames': list(serialize_frames(frame, tb)),
                }
            elif data['type'] == 'GET_FILE':
                filename = data['filename']
                file = ''.join(linecache.getlines(filename))

                response = {
                    'type': 'SET_FILE',
                    'filename': filename,
                    'source': file,
                }
            elif data['type'] == 'SET_PROMPT':
                response = {
                    'type': 'SET_ANSWER',
                    'key': data['key'],
                    'command': data.get('command'),
                    'frame': data.get('frame'),
                    **serialize_answer(
                        data['prompt'], get_frame(frame, data.get("frame"))
                    ),
                }
            elif data['type'] == 'REQUEST_INSPECT':
                response = {
                    'type': 'SET_ANSWER',
                    'key': data['key'],
                    'command': data.get('command'),
                    **serialize_inspect(data['id']),
                }
            elif data['type'] == 'REQUEST_INSPECT_EVAL':
                response = {
                    'type': 'SET_ANSWER',
                    'key': data['key'],
                    'command': data.get('command'),
                    **serialize_inspect_eval(
                        data['prompt'], get_frame(frame, data.get("frame"))
                    ),
                }
            elif data['type'] == 'REQUEST_DIFF_EVAL':
                response = {
                    'type': 'SET_ANSWER',
                    'key': data['key'],
                    'command': data.get('command'),
                    **serialize_diff_eval(
                        data['left'],
                        data['right'],
                        get_frame(frame, data.get("frame")),
                    ),
                }
            elif data['type'] == 'REQUEST_SUGGESTION':
                response = {
                    'type': 'SET_SUGGESTION',
                    **serialize_suggestion(
                        data['prompt'],
                        data['from'],
                        data['to'],
                        data['cursor'],
                        get_frame(frame, data.get("frame")),
                    ),
                }
            elif data['type'] == 'DO_COMMAND':
                command = data['command']
                response = {'type': 'ACK', 'command': command}
                if command == 'run':
                    clear_step()
                    stop_trace(frame)
                elif command == 'stop':
                    clear_step()
                    stop_trace(frame)
                    die()
                else:
                    add_step(command, frame)
                stop = True
            else:
                response = {
                    'type': 'error',
                    'message': f"Unknown type {data['type']}",
                }
            log.debug(f'Got {data} answering with {response}')
            response['local'] = True
            await ws.send_json(response)
            if stop:
                break
        elif msg.type == WSMsgType.ERROR:
            log.error(f'WebSocket closed', exc_info=ws.exception())
            stop_trace(frame)



import os
from pathlib import Path
from subprocess import run

from . import config, run_file, shell

os.environ["PYTHONBREAKPOINT"] = "kalong.breakpoint"
config.from_args()

if config.server:
    from .server import serve

    serve()

elif config.inject:
    kalong_dir = Path(__file__).resolve().parent.parent
    gdb_command = ['gdb', '-p', str(config.inject), '-batch'] + [
        "-eval-command=call %s" % hook
        for hook in [
            '(int) PyGILState_Ensure()',  # Getting the GIL
            '(int) PyRun_SimpleString("'
            f'print(\'* Kalong injection from {os.getpid()} *\');'
            'import sys;'  # Putting kalong project directory in sys path:
            f'sys.path.insert(0, \'{kalong_dir}\');'
            'import kalong;'  # Setting breakpoint:
            'kalong.break_above(2);'
            '")',
            # Releasing the GIL with the PyGILState_Ensure handle:
            '(void) PyGILState_Release($1)',
        ]
    ]
    print(f'Running: {" ".join(gdb_command)}')
    run(gdb_command)

else:
    if config.command:
        run_file(*config.command)
    else:
        shell()



import asyncio
import json
import logging
import os
import signal
from itertools import chain
from pathlib import Path

from aiohttp import WSMsgType, web
from aiohttp.web_runner import GracefulExit

from . import config
from .errors import NoClientFoundError
from .utils import basicConfig, USER_SIGNAL
from .websockets import websocket_options

log = logging.getLogger(__name__)
basicConfig(level=config.log_level)


def maybe_bail(app):
    if not app['front'] and not app['back']:
        log.info('No remaining clients, gracefully bailing.')
        raise GracefulExit()
    else:
        log.info(
            f"Not closing since {len(app['front'])} front and "
            + f"{len(app['back'])} back connections remains."
        )


async def shutdown(app):
    log.info('App shutdown, closing remaining websockets.')
    await asyncio.gather(
        *[
            ws.close()
            for ws in chain(app['front'].values(), app['back'].values())
        ]
    )
    app['front'].clear()
    app['back'].clear()


async def peer(app, side, origin):
    if origin in app[side]:
        return app[side][origin]
    else:
        for delay in [0.01, 0.025, 0.05, 0.1, 0.25, 0.5, 1, 2.5, 5, 10]:
            await asyncio.sleep(delay)
            if origin in app[side]:
                return app[side][origin]
        raise NoClientFoundError()


async def websocket(request):
    side = request.match_info['side']
    other_side = 'back' if side == 'front' else 'front'
    origin = request.match_info['origin']
    ws = web.WebSocketResponse(**websocket_options)

    state = ws.can_prepare(request)
    if not state.ok:
        log.debug(f'Sending {side} app for {origin}')
        return web.FileResponse(
            Path(__file__).parent / 'assets' / 'index.html'
        )

    await ws.prepare(request)

    log.debug(f'{side.title()} app connected for {origin}')
    request.app[side][origin] = ws

    async for msg in ws:
        if msg.type == WSMsgType.TEXT:
            data = json.loads(msg.data)
            log.debug(f'{side} -> {other_side}: {data}')
            pair = await peer(request.app, other_side, origin)

            if data["type"] == "DO_COMMAND":
                if data["command"] in ["pause", "kill"]:
                    pid = int(origin.split("__")[1].split("--")[0])
                    os.kill(
                        pid,
                        signal.SIGTERM
                        if data["command"] == "kill"
                        else USER_SIGNAL,
                    )
                    continue

            await pair.send_json(data)
        elif msg.type == WSMsgType.ERROR:
            log.error(f'{side.title()} closed', exc_info=ws.exception())

    log.debug(f'Closing {side}')
    await ws.close()
    del request.app[side][origin]

    if not config.detached and origin in request.app[other_side]:
        log.debug(f'Closing {other_side} due to {side} closing.')
        await request.app[other_side][origin].close()
        del request.app[other_side][origin]

    maybe_bail(request.app)
    return ws


def serve():
    app = web.Application()
    app['front'] = {}
    app['back'] = {}
    app.on_shutdown.append(shutdown)
    app.router.add_get(r'/{side:(front|back)}/{origin}', websocket)
    app.router.add_static('/assets/', Path(__file__).parent / 'assets')
    web.run_app(app, host=config.host, port=config.port, print=False)



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['kalong', 'kalong.utils', 'kalong.utils.doc_lookup']

package_data = \
{'': ['*'],
 'kalong': ['assets/*',
            'front/*',
            'front/Answer/*',
            'front/Code/*',
            'front/Prompt/*',
            'front/font/*']}

install_requires = \
['aiohttp', 'jedi']

extras_require = \
{'dev': ['black', 'flake8', 'isort'],
 'extras': ['cutter', 'uncompyle6'],
 'test': ['flake8', 'pytest-cov', 'pytest-flake8', 'pytest-isort', 'pytest'],
 'tools': ['sphinx']}

setup(name='kalong',
      version='0.1.1',
      description='A new take on debugging',
      author='Florian Mounier',
      author_email='paradoxxx.zero@gmail.com',
      url='https://github.com/paradoxxxzero/kalong',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
     )



import os
import signal
import socket
import sys
import threading

try:
    from log_colorizer import basicConfig
except ImportError:
    from logging import basicConfig  # noqa


def current_origin():
    return f'{socket.getfqdn()}__{os.getpid()}--{threading.get_ident()}'


class fake_argv:
    def __init__(self, *args):
        self.args = args

    def __enter__(self):
        self._old_args = sys.argv
        sys.argv = self.args

    def __exit__(self, *err):
        sys.argv = self._old_args


USER_SIGNAL = signal.SIGILL if sys.platform == "win32" else signal.SIGUSR1
